

namespace GossamerVR.AutoSave
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;

    using SimpleJSON;
    
    using Debug = UnityEngine.Debug;
    using System.Threading;

    /// <summary>
    /// Allows saving backup files at various rates.
    /// </summary>
    public class AutoSave: MVRScript
    {
        private const string savePath = @"Saves\PluginData\";
        public const string ScriptName = "GVR AutoSave";
        private bool everythingIsTerrible = false;
        private string generalError = "";
        public bool HasError => this.everythingIsTerrible || !string.IsNullOrEmpty(generalError);

        private JSONStorableBool debugBoolJSON;
        private JSONStorableBool fiveMinuteJSON, fifteenMinuteJSON, thirtyMinuteJSON, lastSceneJSON;
        private JSONStorableString instructionsJSON;

        public bool IsDebugMode => this.debugBoolJSON != null && this.debugBoolJSON.val;

        private bool isInitialized;
        private bool isSubscribed;
        private float time;
        private int iterations5m;

        /// <summary>
        /// Re-subscribes to scene events if we are re-activating post-Initialization.
        /// </summary>
        private void OnEnable()
        {            
            this.everythingIsTerrible = false;
            this.generalError = "";

            if (this.isInitialized && !this.isSubscribed)
            {                
                this.Subscribe();        
            }
        }

        /// <summary>
        /// Unsubscribes from scene events and reverts any effects we had.
        /// called automatically by Monobehaviour API when disabled.
        /// </summary>
        private void OnDisable()
        {
            this.DLog("OnDisable()");
            this.Unsubscribe();            
        }

        /// <summary>
        /// Initializes by subscribing to important scene events and 
        /// generating our json items.
        /// </summary>
        public override void Init()
        {            

            this.DLog("Init()");
            base.Init();
            if (containingAtom.name != "CoreControl" || !string.IsNullOrEmpty(containingAtom.uid))
            {
                this.Log("AutoSave must be added as a session plugin.");
                this.generalError += " invalid plugin target, please set as a session plugin !";
                this.StopWorkingAndSulk();
            }
            else
            {
                if (!this.isInitialized)
                {                    
                    this.Subscribe();
                    this.isInitialized = true;
                }
                
                this.debugBoolJSON = new JSONStorableBool("Debug Mode", false);
                this.RegisterBool(this.debugBoolJSON);

                this.fiveMinuteJSON = new JSONStorableBool("5 min save", false);
                this.fifteenMinuteJSON = new JSONStorableBool("15 min save", true);
                this.thirtyMinuteJSON= new JSONStorableBool("30 min save", false);
                //this.lastSceneJSON = new JSONStorableBool("'Last Scene' backup", true, this.OnChangeBackupSettings);

                this.instructionsJSON = new JSONStorableString("Instructions","Each rate of autosave is saved to a different file. " +
                    "All files are stored in <b>Saves/PluginData</b> as autosave_##m.json\n" +
                    "Note that for large files having multiple saves could add latency at the time of saving, as " +
                    "the file will be written multiple times in the same frame.");
            }
        }


        /// <summary>
        /// Initializes the plugin's UI, automatically called by VAM.
        /// </summary>
        public override void InitUI()
        {
            base.InitUI();
            if (this.UITransform != null && this.isInitialized)
            {
                /// LEFT
                this.CreateToggle(this.fiveMinuteJSON, false);
                this.CreateToggle(this.fifteenMinuteJSON, false);
                this.CreateToggle(this.thirtyMinuteJSON, false);
                this.CreateTextField(this.instructionsJSON, false);
                /// RIGHT
                this.CreateToggle(this.debugBoolJSON, true);
                var spacer = this.CreateSpacer(true);
                spacer.height = 10;
                //this.CreateToggle(this.lastSceneJSON, true);


            }
        }

        /// <summary>
        /// Coming soon
        /// </summary>
        /// <param name="value"></param>
        private void OnChangeBackupSettings(bool value)
        {
            this.Log("Backup settings set to " + value.ToString());
        }


        // at some point we might subscribe to vam events but for now we stub it.
        private void Subscribe() => this.isSubscribed = true;
        private void Unsubscribe() => this.isSubscribed = false;

        void Update()
        {                       
            if (!this.everythingIsTerrible && this.isInitialized && (this.fiveMinuteJSON.val || this.fifteenMinuteJSON.val || this.thirtyMinuteJSON.val))
            {
                this.time += Time.deltaTime;
                if (this.time >= 300f)
                {
                    this.time -= 300f;
                    this.iterations5m++;
                    JSONClass saveData = null;                    
                    try
                    {
                        if (this.fiveMinuteJSON.val)
                        {
                            saveData = saveData ?? SuperController.singleton.GetSaveJSON();
                            this.DLog("saving autosave_05m");
                            SuperController.singleton.SaveJSON(saveData, savePath + "autosave_05m.json");
                        }
                        if (this.fifteenMinuteJSON.val && this.iterations5m % 3 == 0)
                        {
                            this.DLog("saving autosave_15m");
                            saveData = saveData ?? SuperController.singleton.GetSaveJSON();
                            SuperController.singleton.SaveJSON(saveData, savePath + "autosave_15m.json");
                        }
                        if (this.thirtyMinuteJSON.val && this.iterations5m % 6 == 0)
                        {
                            this.DLog("saving autosave_30m");
                            saveData = saveData ?? SuperController.singleton.GetSaveJSON();
                            SuperController.singleton.SaveJSON(saveData, savePath + "autosave_30m.json");
                        }
                    }
                    catch (Exception e)
                    {
                        this.Log("Couldn't autosave: "+e.Message, error: true);
                    }
                }
            }
        }


        /// <summary>
        /// Sets the flag showing that something is critically wrong,
        /// and disables the script. This is called when we encounter an issue
        /// that would prevent the plugin from working, such as missing transforms.        
        /// </summary>
        private void StopWorkingAndSulk()
        {
            if (!this.everythingIsTerrible)
            {

                this.Log($"{ScriptName} is shutting itself down with these errors: {this.generalError} \nThings didn't go as planned, and {ScriptName} has decided to sulk. " +
                    "You can try re-enabling or reloading it if you think it's a fluke.", error: true);

                this.everythingIsTerrible = true;
                this.generalError = "";

                // we don't want to just disable the script, we want to show it as disabled
                // in VAM's plugin UI by turning off the toggle. Though I feel like I'm doing this wrong.         
                if (this.enabledJSON != null && this.enabledJSON.toggle != null)
                {
                    this.enabledJSON.toggle.isOn = false;
                }


            }
        }

        private void DLog(string str, bool error = false)
        {
            this.LogIf(this.IsDebugMode, str, error);
        }

        private void LogIf(bool condition, string str, bool error = false)
        {
            if (condition) this.Log(str);
        }

        private void Log(string str, bool error = false)
        {
            if (error) SuperController.LogError("GVR AutoSave: "+str);
            else SuperController.LogMessage("GVR AutoSave: "+str);
        }



    }
}